import base64
import logging

logger = logging.getLogger(__name__)

import time
from contextlib import contextmanager

@contextmanager
def timer(name):
    start = time.time()
    yield
    logger.info(f'[{name}] done in {time.time() - start:.2f} s')

def encode_image(file: str) -> str:
    """将图片编码为 Base64"""
    with timer(f'encode [{file}] to base64'):
        with open(file, 'rb') as f:
            return base64.encodebytes(f.read()).decode('utf8').replace('\n', '').replace('\r', '')

import subprocess
import os
def runcmd(command):
    ret = subprocess.run(command,shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE,encoding="utf-8",timeout=1)
    if ret.returncode == 0:
        # print("success:",ret)
        return ret
    else:
        raise ValueError("error: "+ str(ret))

def check_pngquant_bin():
    try:
        runcmd('pngquant --help')
    except Exception as e:
        print('error:', e)
        print('用于图片压缩的工具 pngquant 不存在，请先在 https://pngquant.org/ 上下载并安装！！！')
        os._exit(1)

def compress_image_by_zipflipng(src: str):
    from zopflipng import png_optimize

    with open(src, 'rb') as fp:
        data = fp.read()
    result, code = png_optimize(data)
    # if code ==0 ,png compression success
    if code == 0:
        # save png
        with open(src, 'wb') as f:
            f.write(result)
            f.close()
    else:
        logger.error('optimize by zopfilong failed, ignore this image')

def png2webp(src: str, quality=30):
    """ PNG 转 WEBP
    ---
    目前是压缩最佳方式
    """
    with timer('convert png to webp'):
        from PIL import Image
        im = Image.open(src)
        dst = src.replace('.png', ".webp")
        im.save(dst, "WEBP", quality=quality)
        return dst

def png2avif(src: str, quality: int=30):
    """目前最新的图片格式，据说比 webp 更轻
    但考虑到兼容性，目前还是暂时使用 webp
    ---
    经过测试，极限状态下比 webp 更强，可以用于缩略图
    """
    from PIL import Image
    import pillow_avif
    with timer('convert png to avif'):
        im = Image.open(src)
        dst = src.replace('.png', ".avif")
        im.save(dst,'AVIF', quality=quality)
    return dst

def compress_image(file: str, zop_enable=False) -> str:
    """压缩图片，返回图片路径
    pngquant: 有损压缩，压缩率高，图片会失真，但速度快
    zopflipng: 无损压缩，压缩率低，图片不失真，但速度慢，适合后台慢慢运行
    
    ---
    经过比对，不通过压缩，直接将 png 转变为 webp, 能同时保持图片清晰度及最大压缩率
    """
    import pngquant
    dst = file.replace('.png', '.compressed.png')
    import time
    with timer('Image compress by pngquant'):
        pngquant.quant_image(file, dst=dst)
    if zop_enable:
        with timer('Image decompress by zopflipng'):
            compress_image_by_zipflipng(dst)
    return dst

def make_thumbnail(image: str, quality=10) -> str:
    """ 制作缩略图 """
    return 'data:image/cvif;base64, ' + encode_image(png2avif(image, quality=quality))

from PIL import Image

def handle_image(image: Image.Image, webp_quality=50, avif_quality=10):
    from io import BytesIO
    import base64
    # to webp
    webp_buf = BytesIO()
    image.save(webp_buf, 'WEBP', quality=webp_quality)
    # to avif (thumbnail)
    import pillow_avif as _
    avif_buf = BytesIO()
    image.save(avif_buf, 'AVIF', quality=avif_quality)
    # to base64
    def to_base64(b: BytesIO):
        return base64.b64encode(b.getvalue()).decode('utf8').replace('\n', '').replace('\r', '')
    result = to_base64(webp_buf)
    suffix = 'webp'
    thumbnail = 'data:image/cvif;base64, ' + to_base64(avif_buf)
    return (
        result,
        suffix,
        thumbnail
    )
    

if __name__ == '__main__':
    LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
    check_pngquant_bin()
    # print(encode_image('assets/demo.png'))
    # dst = compress_image('assets/demo.png', zop_enable=True)
    png2webp('assets/demo.png', quality=1)
    def test_avif_to_base64():
        """ 测试 avif 图片转 base64 后，文件大小"""
        avif = png2avif('assets/demo.png', quality=10)
        encoded = encode_image(avif)
        print(f'{len(encoded)/1024} kB')
    test_avif_to_base64()



# -*- coding: utf-8 -*-
"""
A simple GraphQL client that works over Websocket as the transport
protocol, instead of HTTP.
This follows the Apollo protocol.
https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
"""

import json
import threading
import uuid
import queue
import logging
from typing import Callable
import time

import websocket


GQL_WS_SUBPROTOCOL = "graphql-ws"

# all the message types
GQL_CONNECTION_INIT = 'connection_init'
GQL_START = 'start'
GQL_STOP = 'stop'
GQL_CONNECTION_TERMINATE = 'connection_terminate'
GQL_CONNECTION_ERROR = 'connection_error'
GQL_CONNECTION_ACK = 'connection_ack'
GQL_DATA = 'data'
GQL_ERROR = 'error'
GQL_COMPLETE = 'complete'
GQL_CONNECTION_KEEP_ALIVE = 'ka'

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ConnectionException(Exception):
    """Exception thrown during connection errors to the GraphQL server"""

class InvalidPayloadException(Exception):
    """Exception thrown if payload recived from server is mal-formed or cannot be parsed """

class GraphQLClient():
    """
    A simple GraphQL client that works over Websocket as the transport
    protocol, instead of HTTP.
    This follows the Apollo protocol.
    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """
    def __init__(self, url):
        self.ws_url = url
        self._connection_init_done = False
        # cache of the headers for a session
        self._headers = None
        # map of subscriber id to a callback function
        self._subscriber_callbacks = {}
        # our general receive queue
        self._queue = queue.Queue()
        # map of queues for each subscriber
        self._subscriber_queues = {}
        self._shutdown_receiver = False
        self._subscriptions = []
        self._closed = False
        self.connect()

    def connect(self) -> None:
        """
        Initializes a connection with the server.
        """
        self._connection = websocket.create_connection(self.ws_url,
                                                       subprotocols=[GQL_WS_SUBPROTOCOL])
        # start the reciever thread
        self._recevier_thread = threading.Thread(target=self._receiver_task)
        self._recevier_thread.start()

    def _reconnect(self):
        subscriptions = self._subscriptions
        self.__init__(self.ws_url)

        for subscription in subscriptions:
            self.subscribe(query=subscription['query'],
                           variables=subscription['variables'],
                           headers=subscription['headers'],
                           callback=subscription['callback'])

    def __dump_queues(self):
        logger.debug('[GQL_CLIENT] => Dump of all the internal queues')
        logger.debug('[GQL_CLIENT] => Global queue => \n %s', self._queue.queue)
        dumps = list(map(lambda q: (q[0], q[1].queue), self._subscriber_queues.items()))
        logger.debug('[GQL_CLIENT] => Operation queues: \n %s', dumps)

    # wait for any valid message, while ignoring GQL_CONNECTION_KEEP_ALIVE
    def _receiver_task(self):
        """the recieve function of the client. Which validates response from the
        server and queues data """
        reconnected = False
        while not self._shutdown_receiver and not reconnected:
            self.__dump_queues()
            try:
                res = self._connection.recv()
            except websocket._exceptions.WebSocketConnectionClosedException as e:
                # TODO 此处的 reconnect 有问题，注释掉
                # self._reconnect()
                # reconnected = True
                # continue
                # 直接关闭，通知上层，由上层处理
                self._shutdown_receiver = True
                self._connection.close()
                self._closed = True
                break
            try:
                msg = json.loads(res)
            except json.JSONDecodeError as err:
                logger.warning('Ignoring. Server sent invalid JSON data: %s \n %s', res, err)
                continue

            # ignore messages which are GQL_CONNECTION_KEEP_ALIVE
            if msg['type'] != GQL_CONNECTION_KEEP_ALIVE:

                # check all GQL_DATA and GQL_COMPLETE should have 'id'.
                # Otherwise, server is sending malformed responses, error out!
                if msg['type'] in [GQL_DATA, GQL_COMPLETE] and 'id' not in msg:
                    # TODO: main thread can't catch this exception; setup
                    # exception queues. but this scenario will only happen with
                    # servers having glaring holes in implementing the protocol
                    # correctly, which is rare. hence this is not very urgent
                    err = f'Protocol Violation.\nExpected "id" in {msg}, but could not find.'
                    raise InvalidPayloadException(err)

                # if the message has an id, it is meant for a particular operation
                if 'id' in msg:
                    op_id = msg['id']

                    # put it in the correct operation/subscriber queue
                    if op_id in self._subscriber_queues:
                        self._subscriber_queues[op_id].put(msg)
                    else:
                        self._subscriber_queues[op_id] = queue.Queue()
                        self._subscriber_queues[op_id].put(msg)

                    # if a callback fn exists with the id, call it
                    if op_id in self._subscriber_callbacks:
                        user_fn = self._subscriber_callbacks[op_id]
                        user_fn(op_id, msg)

                # if it doesn't have an id, put in the global queue
                else:
                    self._queue.put(msg)

    def _insert_subscriber(self, op_id, callback_fn):
        self._subscriber_callbacks[op_id] = callback_fn

    def _remove_subscriber(self, op_id):
        del self._subscriber_callbacks[op_id]

    def _create_operation_queue(self, op_id):
        self._subscriber_queues[op_id] = queue.Queue()

    def _remove_operation_queue(self, op_id):
        if op_id in self._subscriber_queues:
            del self._subscriber_queues[op_id]

    def _get_operation_result(self, op_id):
        return self._subscriber_queues[op_id].get()

    def _connection_init(self, headers=None):
        # if we have already initialized and the passed headers are same as
        # prev headers, then do nothing and return
        if self._connection_init_done and headers == self._headers:
            return

        self._headers = headers
        # send the `connection_init` message with the payload
        payload = {'type': GQL_CONNECTION_INIT, 'payload': {'headers': headers}}
        self._connection.send(json.dumps(payload))

        res = self._queue.get()

        if res['type'] == GQL_CONNECTION_ERROR:
            err = res['payload'] if 'payload' in res else 'unknown error'
            raise ConnectionException(err)
        if res['type'] == GQL_CONNECTION_ACK:
            self._connection_init_done = True
            return

        err_msg = "Unknown message from server, this client did not understand. " + \
            "Original message: " + res['type']
        raise ConnectionException(err_msg)

    def _start(self, payload, callback=None):
        """ pass a callback function only if this is a subscription """
        op_id = uuid.uuid4().hex
        frame = {'id': op_id, 'type': GQL_START, 'payload': payload}
        self._create_operation_queue(op_id)
        if callback:
            self._insert_subscriber(op_id, callback)
        self._connection.send(json.dumps(frame))
        return op_id

    def _stop(self, op_id):
        payload = {'id': op_id, 'type': GQL_STOP}
        self._connection.send(json.dumps(payload))

    def query(self, query: str, variables: dict = None, headers: dict = None) -> dict:
        """
        Run a GraphQL query or mutation. The `query` argument is a GraphQL query
        string. You can pass optional variables and headers.

        PS: To run a subscription, see the `subscribe` method.
        """
        self._connection_init(headers)
        payload = {'headers': headers, 'query': query, 'variables': variables}
        op_id = self._start(payload)
        res = self._get_operation_result(op_id)
        self._stop(op_id)
        ack = self._get_operation_result(op_id)
        if ack['type'] != GQL_COMPLETE:
            logger.warning('Expected to recieve complete, but received: %s', ack)
        self._remove_operation_queue(op_id)
        return res

    def subscribe(self, query: str, variables: dict = None, headers: dict = None,
                  callback: Callable[[str, dict], None] = None) -> str:
        """
        Run a GraphQL subscription.

        Parameters:
        query (str): the GraphQL query string
        callback (function): a callback function. This is mandatory.
        This callback function is called, everytime there is new data from the
        subscription.
        variables (dict): (optional) GraphQL variables
        headers (dict): (optional) a dictionary of headers for the session

        Returns:
        op_id (str): The operation id (a UUIDv4) for this subscription operation
        """

        # sanity check that the user passed a valid function
        if not callback or not callable(callback):
            raise TypeError('the argument `callback` is mandatory and it should be a function')

        self._connection_init(headers)
        payload = {'headers': headers, 'query': query, 'variables': variables}
        op_id = self._start(payload, callback)
        self._subscriptions.append({
            'query': query,
            'variables': variables,
            'headers': headers,
            'callback': callback
        })
        return op_id

    def stop_subscribe(self, op_id: str) -> None:
        """
        Stop a subscription. Takes an operation ID (`op_id`) and stops the
        subscription.
        """
        self._stop(op_id)
        self._remove_subscriber(op_id)
        self._remove_operation_queue(op_id)

    def close(self) -> None:
        """
        Close the connection with the server. To reconnect, use the `connect`
        method.
        """
        self._shutdown_receiver = True
        self._recevier_thread.join()
        self._connection.close()
        self._closed = True

    def __enter__(self):
        """ enter method for context manager """
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """ exit method for context manager """
        self.close()

    def run(self):
        while not self._closed:
            time.sleep(1)



import setuptools
import os
import io

long_description = 'Aiy worker'
  
if os.path.exists("requirements.txt"):
    install_requires = [i for i in io.open("requirements.txt").read().split("\n") if i]
else:
    install_requires = []
print(install_requires)
setuptools.setup(
    name="aiy_worker",
    version="0.0.6",
    author="zgljl2012",
    # license = 'MIT License',
    # author_email="pengshiyuyx@gmail.com",
    description="aiy worker",
    long_description=long_description,
    # long_description_content_type="text/x-rst",
    # url="https://github.com/mouday/chinesename",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires = install_requires,
    # include_package_data=True,  # 自动打包文件夹内所有数据
    # 如果需要包含多个文件可以单独配置 MANIFEST.in
    package_data = {
            # If any package contains *.txt or *.rst files, include them:
            'aiy_worker': [],
    },
    # 如果需要支持脚本方法运行，可以配置入口点
    # entry_points={
    #     'console_scripts': [
    #         'chinesename = chinesename.run:main'
    #     ]
    # }
)
